//
//  SAViewController.m
//  SocialAuth
//
//  Created by Dan Fleming on 10/26/12.
//  Copyright (c) 2012 Dan Fleming. All rights reserved.
//

#import "SAViewController.h"
#import "SACategories.h"
#import "GCOAuth.h"

@interface SAViewController ()
@property(nonatomic,assign) BOOL wasDismissed;
-(void) didSucceed:(NSDictionary*)data;
-(void) didFail:(NSError*)error;
-(void) didCancel;
@end

@implementation SAViewController
@synthesize config = _config;
@synthesize delegate = _delegate;
@synthesize dismissalBlock = _dismissalBlock;

#pragma mark -
#pragma mark View lifecycle

-(id) initWithConfig:(NSDictionary*)config delegate:(id<SAAuthenticationDelegate>)delegate {
    self = [super initWithNibName:nil bundle:nil];
    if(self){
        _config = config;
        self.delegate = delegate;
        self.title = config[kSAConfigServiceTitle];
    }
    return self;
}

-(id) initWithConfig:(NSDictionary*)config {
    self = [self initWithConfig:config delegate:nil];
    return self;
}

-(void) viewWillAppear:(BOOL)animated {
    [super viewWillAppear:animated];
    if(self.navigationController && self.navigationController.viewControllers.count == 1){
        // Looks like we're presented in a modal
        if(self.navigationItem.rightBarButtonItem == nil && self.navigationItem.leftBarButtonItem == nil){
            // If no close button has been placed on this controller, add the default close button
            self.navigationItem.rightBarButtonItem = [[UIBarButtonItem alloc] initWithTitle:@"Cancel"
                                                                                      style:UIBarButtonItemStyleDone
                                                                                     target:self
                                                                                     action:@selector(cancelAuthentication)];
        }
    }
}

-(void) viewWillDisappear:(BOOL)animated {
    [super viewWillDisappear:animated];
    if(!self.wasDismissed){
        self.wasDismissed = YES;
        // If not explicitly dismissed, the user cancelled (back button, etc)
        [self didCancel];
    }
}

#pragma mark -
#pragma mark Call delegate, dismiss window

-(void) dismiss {
    if(!self.wasDismissed){
        self.wasDismissed = YES;
        if(self.dismissalBlock){
            // If the user specified a custom dismissal block, call it
            self.dismissalBlock();
        }else{
            // If not, take a guess how to get out of this mess
            if(self.navigationController){
                if(self.navigationController.viewControllers.count > 1){
                    // If there is more than one viewcontroller on the stack, we're probably safe to just pop
                    [self.navigationController popViewControllerAnimated:YES];
                }else{
                    // If not, we're probably presented in a modal with a navigation controller wrapper
                    [self.navigationController dismissModalViewControllerAnimated:YES];
                }
            }
        }
    }
}

-(void) didSucceed:(NSDictionary*)data {
    if(self.delegate && [self.delegate respondsToSelector:@selector(clientAuthentication:didSucceed:)]){
        [self.delegate clientAuthentication:self.config didSucceed:data];
    }
    [self dismiss];
}

-(void) didFail:(NSError*)error {
    if(self.delegate && [self.delegate respondsToSelector:@selector(clientAuthentication:didFail:)]){
        [self.delegate clientAuthentication:self.config didFail:error];
    }
    [self dismiss];
}

-(void) didCancel {
    if(self.delegate && [self.delegate respondsToSelector:@selector(clientAuthenticationWasCancelled:)]){
        [self.delegate clientAuthenticationWasCancelled:self.config];
    }
    [self dismiss];
}

-(void) cancelAuthentication {
    [self didCancel];
}

@end



@interface SAWebViewController ()
@property(nonatomic,strong) UIActivityIndicatorView *activityView;
@property(nonatomic,assign) BOOL isLoading;
-(void) loadInitialRequest;
-(void) hideWebView;
@end

@implementation SAWebViewController
@synthesize webView = _webView;
@synthesize activityView = _activityView;
@synthesize isLoading = _isLoading;

#pragma mark -
#pragma mark View lifecycle

- (void)viewDidLoad {
    [super viewDidLoad];
    self.view.backgroundColor = [UIColor whiteColor];
    
    // Add an activity indicator for the first load, center in frame
    _activityView = [[UIActivityIndicatorView alloc] initWithActivityIndicatorStyle:UIActivityIndicatorViewStyleGray];
    CGSize ws = self.view.bounds.size;
    CGSize vs = _activityView.bounds.size;
    _activityView.frame = CGRectMake(floorf((ws.width - vs.width) / 2.f),
                                     floorf((ws.height - vs.height) / 2.f),
                                     vs.width, vs.height);
    _activityView.autoresizingMask = (UIViewAutoresizingFlexibleLeftMargin | UIViewAutoresizingFlexibleRightMargin | UIViewAutoresizingFlexibleTopMargin | UIViewAutoresizingFlexibleBottomMargin);
    [self.view addSubview:_activityView];
    
    // Add the web view
    _webView = [[UIWebView alloc] initWithFrame:self.view.bounds];
    _webView.autoresizingMask = (UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight);
    _webView.delegate = self;
    _webView.hidden = YES;
    [self.view addSubview:_webView];
}

-(void) viewDidAppear:(BOOL)animated {
    [super viewDidAppear:animated];
    if(_webView.request == nil){
        [self loadInitialRequest];
    }
}

-(void) viewWillDisappear:(BOOL)animated {
    [_webView stopLoading];
    self.isLoading = NO;
    [super viewWillDisappear:animated];
}

#pragma mark -
#pragma mark UIWebViewDelegate methods

-(void) webViewDidStartLoad:(UIWebView *)webView {
    self.isLoading = YES;
}

-(void) webViewDidFinishLoad:(UIWebView *)webView {
    self.isLoading = NO;
    if(_webView.hidden){
        _webView.alpha = 0;
        _webView.hidden = NO;
        [UIView animateWithDuration:0.35f animations:^(){
            _webView.alpha = 1.f;
        } completion:^(BOOL completed){
        }];
    }
}

-(void) hideWebView {
    if(!_webView.hidden){
        _webView.alpha = 1.f;
        [UIView animateWithDuration:0.35f animations:^(){
            _webView.alpha = 0.f;
        } completion:^(BOOL completed){
            _webView.hidden = YES;
        }];
    }
}

#pragma mark -
#pragma mark The rest

-(void) setIsLoading:(BOOL)isLoading {
    if(_isLoading == isLoading) return;
    _isLoading = isLoading;
    [[UIApplication sharedApplication] setNetworkActivityIndicatorVisible:_isLoading];
    if(_isLoading){
        if(!_activityView.isAnimating)
            [_activityView startAnimating];
    }else{
        if(_activityView.isAnimating)
            [_activityView stopAnimating];
    }
}

-(void) loadInitialRequest { }

@end


@interface SAOAuth1WebViewController()
@property(nonatomic,strong) NSOperationQueue *opQueue;
@property(nonatomic,strong) NSString *requestOAuthToken;
@property(nonatomic,strong) NSString *requestOAuthTokenSecret;
@end

@implementation SAOAuth1WebViewController
@synthesize opQueue, requestOAuthToken, requestOAuthTokenSecret;


-(void) loadInitialRequest {
    // Clear cookies for this domain
    NSString *redirectURL = self.config[kSAConfigWebAuthenticationURL];
    [SAUtil deleteHTTPCookiesWhereHostContains:self.config[kSAConfigWebBaseDomain]];
    [SAUtil deleteHTTPCookiesWhereHostContains:[[NSURL URLWithString:redirectURL] host]];
    
    // Build the request token request
    self.isLoading = YES;
    NSURL *url = [NSURL URLWithString:self.config[kSAConfigOAuth1RequestTokenURL]];
    NSURLRequest *req = [GCOAuth URLRequestForPath:url.path
                                    POSTParameters:nil
                                            scheme:url.scheme
                                              host:url.hostAndPort
                                       consumerKey:self.config[kSAConfigOAuth1ConsumerKey]
                                    consumerSecret:self.config[kSAConfigOAuth1ConsumerSecret]
                                       accessToken:nil
                                       tokenSecret:nil];
    
    self.requestOAuthToken = nil;
    self.requestOAuthTokenSecret = nil;
    self.opQueue = [[NSOperationQueue alloc] init];
    __weak SAOAuth1WebViewController *blockSelf = self;
    
    // Get request token
    [NSURLConnection sendAsynchronousRequest:req queue:self.opQueue completionHandler:^(NSURLResponse *response, NSData *data, NSError *error) {
        if(error){
            dispatch_async(dispatch_get_main_queue(), ^{
                [blockSelf didFail:error];
            });
        }else{
            NSHTTPURLResponse *httpResponse = (NSHTTPURLResponse*)response;
            if([httpResponse statusCode] >= 200 && [httpResponse statusCode] < 300){
                // Encode response body as string and parse
                NSString *responseString = [[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding];
                NSDictionary *responseParams = [NSDictionary dictionaryWithURLEncodedString:responseString];
                
                // Save the request token and secret
                blockSelf.requestOAuthToken = [responseParams objectForKey:@"oauth_token"];
                blockSelf.requestOAuthTokenSecret = [responseParams objectForKey:@"oauth_token_secret"];
                
                if(blockSelf.requestOAuthToken && blockSelf.requestOAuthTokenSecret){
                    // Build authorization URL and load in web view
                    NSDictionary *urlParams = @{
                        @"oauth_token": blockSelf.requestOAuthToken,
                        @"oauth_callback": blockSelf.config[kSAConfigOAuth1RedirectURL]
                    };
                    NSURL *authUrl = [NSURL URLWithString:[blockSelf.config[kSAConfigWebAuthenticationURL] stringByAppendingQueryParameters:urlParams]];
                    dispatch_async(dispatch_get_main_queue(), ^{
                        [blockSelf.webView loadRequest:[NSURLRequest requestWithURL:authUrl]];
                    });
                }else{
                    // TODO: better error - did not receive request token
                    error = [NSError errorWithDomain:@"" code:0 userInfo:nil];
                    dispatch_async(dispatch_get_main_queue(), ^{
                        [blockSelf didFail:error];
                    });
                }
            }else{
                // TODO: better error - non-successful status code
                error = [NSError errorWithDomain:@"" code:0 userInfo:nil];
                dispatch_async(dispatch_get_main_queue(), ^{
                    [blockSelf didFail:error];
                });
            }
        }
    }];
}

- (BOOL)webView:(UIWebView*)webView shouldStartLoadWithRequest:(NSURLRequest*)request navigationType:(UIWebViewNavigationType)navigationType {
    NSString *toMatch = self.config[kSAConfigOAuth1RedirectURL];
    NSString *urlString = request.URL.absoluteString;
    if([urlString hasPrefix:toMatch]){
        
        // Look for oauth_verifier
        NSDictionary *qp = [NSDictionary dictionaryWithURLEncodedString:request.URL.query];
        NSDictionary *pp = qp[@"oauth_verifier"] ? @{@"oauth_verifier": qp[@"oauth_verifier"]} : nil;
        
        // Build the access token request
        NSURL *url = [NSURL URLWithString:self.config[kSAConfigOAuth1AccessTokenURL]];
        NSURLRequest *req = [GCOAuth URLRequestForPath:url.path
                                        POSTParameters:pp
                                                scheme:url.scheme
                                                  host:url.hostAndPort
                                           consumerKey:self.config[kSAConfigOAuth1ConsumerKey]
                                        consumerSecret:self.config[kSAConfigOAuth1ConsumerSecret]
                                           accessToken:self.requestOAuthToken
                                           tokenSecret:self.requestOAuthTokenSecret];
        
        __weak SAOAuth1WebViewController *blockSelf = self;
        [self hideWebView];
        self.isLoading = YES;
        
        [NSURLConnection sendAsynchronousRequest:req queue:self.opQueue completionHandler:^(NSURLResponse *response, NSData *data, NSError *error) {
            // TODO: check status code
            if(error){
                dispatch_async(dispatch_get_main_queue(), ^{
                    [blockSelf didFail:error];
                });
            }else{
                NSHTTPURLResponse *httpResponse = (NSHTTPURLResponse*)response;
                if([httpResponse statusCode] >= 200 && [httpResponse statusCode] < 300){
                    // Encode response body as string and parse
                    NSString *responseString = [[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding];
                    NSDictionary *responseParams = [NSDictionary dictionaryWithURLEncodedString:responseString];
                    
                    NSString *token = responseParams[@"oauth_token"];
                    NSString *secret = responseParams[@"oauth_token_secret"];
                    
                    if(token && secret){
                        // TODO: save additional returned values? (like screen_name / user_id from twitter)
                        dispatch_async(dispatch_get_main_queue(), ^{
                            [blockSelf didSucceed:@{
                                kSADataOAuth1AccessToken: token,
                                kSADataOAuth1TokenSecret: secret
                            }];
                        });
                    }else{
                        // TODO: better error - did not receive access token
                        error = [NSError errorWithDomain:@"" code:0 userInfo:nil];
                        dispatch_async(dispatch_get_main_queue(), ^{
                            [blockSelf didFail:error];
                        });
                    }
                }else{
                    // TODO: better error - non-successful status code
                    error = [NSError errorWithDomain:@"" code:0 userInfo:nil];
                    dispatch_async(dispatch_get_main_queue(), ^{
                        [blockSelf didFail:error];
                    });
                }
                
            }
        }];
        return NO;
    }
    return YES;
}

@end



@implementation SAOAuth2WebViewController

-(void) loadInitialRequest {
    NSDictionary *params = @{
        @"client_id": self.config[kSAConfigOAuth2ClientID],
        @"redirect_uri": self.config[kSAConfigOAuth2RedirectURL],
        @"response_type": @"token"
    };
    
    // Clear cookies for this domain
    NSString *redirectURL = self.config[kSAConfigWebAuthenticationURL];
    [SAUtil deleteHTTPCookiesWhereHostContains:self.config[kSAConfigWebBaseDomain]];
    [SAUtil deleteHTTPCookiesWhereHostContains:[[NSURL URLWithString:redirectURL] host]];
    
    NSString *url = [redirectURL stringByAppendingQueryParameters:params];
    NSURLRequest *req = [NSURLRequest requestWithURL:[NSURL URLWithString:url]];
    [self.webView loadRequest:req];
}

- (BOOL)webView:(UIWebView*)webView shouldStartLoadWithRequest:(NSURLRequest*)request navigationType:(UIWebViewNavigationType)navigationType {
    NSString *toMatch = self.config[kSAConfigOAuth2RedirectURL];
    NSString *urlString = request.URL.absoluteString;
    if([urlString hasPrefix:toMatch]){
        // Get parameters from URL fragment (response_type = token)
        NSDictionary *params = [NSDictionary dictionaryWithURLEncodedString:request.URL.fragment];
        
        // Fallback to (response_type = code) for errors and sites that don't support token
        if([params count] == 0) params = [NSDictionary dictionaryWithURLEncodedString:request.URL.query];
        
        // Capture the error if it exists
        NSString *errorCode = [params valueForKey:@"error"];
        NSString *errorReason = [params valueForKey:@"error_reason"];
        
        // Get the access token
        NSString *token = [params valueForKey:@"access_token"];
        if(token == nil) token = [params valueForKey:@"code"];
        
        if(errorCode || [token length] == 0){
            // This means a user denied an app access to their data
            if([@"user_denied" isEqualToString:errorReason]){
                [self didCancel];
            }else{
                // TODO: output better error
                [self didFail:[NSError errorWithDomain:@"" code:0 userInfo:nil]];
            }
        }else{ 
            [self didSucceed:@{kSADataOAuth2AccessToken: token}];
        }
        return NO;
    }
    return YES;
}

@end

